{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# 目标检测输出层"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 1,
   "metadata": {},
   "outputs": [],
   "source": [
    "import set_env\n",
    "from d2py.utils.file import mkdir\n",
    "root_dir = \".temp\"\n",
    "mkdir(f\"{root_dir}/logs\")\n",
    "mkdir(f\"{root_dir}/libs\")"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "配置信息："
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "metadata": {},
   "outputs": [],
   "source": [
    "from dataclasses import dataclass\n",
    "\n",
    "@dataclass\n",
    "class Config:\n",
    "    shape: tuple # 输入数据 shape\n",
    "    name: str = \"data\" # 输入数据名称\n",
    "    dtype: str = \"float32\" # 输入数据类型"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## 构建 PyTorch 人脸检测模型"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 3,
   "metadata": {},
   "outputs": [],
   "source": [
    "import torch\n",
    "from torch import nn\n",
    "\n",
    "class M(nn.Module):\n",
    "    def __init__(self):\n",
    "        super().__init__()\n",
    "        self.conv = nn.Conv2d(3, 64, 1, 1, 0, bias=False, groups=1)\n",
    "        self.conv0 = nn.Conv2d(64, 36, 1, 1, 0, bias=False, groups=1)\n",
    "        self.conv00 = nn.Conv2d(64, 36, 1, 1, 0, bias=False, groups=1)\n",
    "        self.resize_2 = nn.Conv2d(64, 64, 1, 2, 0, bias=False, groups=1)\n",
    "        self.conv1 = nn.Conv2d(64, 36, 1, 1, 0, bias=False, groups=1)\n",
    "        self.conv11 = nn.Conv2d(64, 36, 1, 1, 0, bias=False, groups=1)\n",
    "        self.resize_4 = nn.Conv2d(64, 64, 1, 4, 0, bias=False, groups=1)\n",
    "        self.conv2 = nn.Conv2d(64, 48, 1, 1, 0, bias=False, groups=1)\n",
    "        self.conv22 = nn.Conv2d(64, 48, 1, 1, 0, bias=False, groups=1)\n",
    "\n",
    "    def forward(self, x):\n",
    "        _x = self.conv(x)\n",
    "        __x = self.resize_2(_x)\n",
    "        ___x = self.resize_4(_x)\n",
    "        x0 = self.conv0(_x).permute(0, 2, 3, 1) # NCHW => NHWC\n",
    "        x1 = self.conv1(__x).permute(0, 2, 3, 1)\n",
    "        x2 = self.conv2(___x).permute(0, 2, 3, 1)\n",
    "        x0 = x0.reshape(1, -1, 3)\n",
    "        x1 = x1.reshape(1, -1, 3)\n",
    "        x2 = x2.reshape(1, -1, 3)\n",
    "        x = torch.concat((x0, x1, x2), dim=1)\n",
    "\n",
    "        x00 = self.conv00(_x).permute(0, 2, 3, 1) # NCHW => NHWC\n",
    "        x11 = self.conv11(__x).permute(0, 2, 3, 1)\n",
    "        x22 = self.conv22(___x).permute(0, 2, 3, 1)\n",
    "        x00 = x00.reshape(1, -1, 4)\n",
    "        x11 = x11.reshape(1, -1, 4)\n",
    "        x22 = x22.reshape(1, -1, 4)\n",
    "        xx = torch.concat((x00, x11, x22), dim=1)\n",
    "        return torch.softmax(x, dim=2), xx"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## 导出 ONNX 模型"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 4,
   "metadata": {
    "tags": [
     "hidden-output"
    ]
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "Exported graph: graph(%data : Float(1, 3, 48, 80, strides=[11520, 3840, 80, 1], requires_grad=0, device=cpu),\n",
      "      %conv.weight : Float(64, 3, 1, 1, strides=[3, 1, 1, 1], requires_grad=1, device=cpu),\n",
      "      %conv0.weight : Float(36, 64, 1, 1, strides=[64, 1, 1, 1], requires_grad=1, device=cpu),\n",
      "      %conv00.weight : Float(36, 64, 1, 1, strides=[64, 1, 1, 1], requires_grad=1, device=cpu),\n",
      "      %resize_2.weight : Float(64, 64, 1, 1, strides=[64, 1, 1, 1], requires_grad=1, device=cpu),\n",
      "      %conv1.weight : Float(36, 64, 1, 1, strides=[64, 1, 1, 1], requires_grad=1, device=cpu),\n",
      "      %conv11.weight : Float(36, 64, 1, 1, strides=[64, 1, 1, 1], requires_grad=1, device=cpu),\n",
      "      %resize_4.weight : Float(64, 64, 1, 1, strides=[64, 1, 1, 1], requires_grad=1, device=cpu),\n",
      "      %conv2.weight : Float(48, 64, 1, 1, strides=[64, 1, 1, 1], requires_grad=1, device=cpu),\n",
      "      %conv22.weight : Float(48, 64, 1, 1, strides=[64, 1, 1, 1], requires_grad=1, device=cpu)):\n",
      "  %/conv/Conv_output_0 : Float(1, 64, 48, 80, strides=[245760, 3840, 80, 1], requires_grad=0, device=cpu) = onnx::Conv[dilations=[1, 1], group=1, kernel_shape=[1, 1], pads=[0, 0, 0, 0], strides=[1, 1], onnx_name=\"/conv/Conv\"](%data, %conv.weight), scope: __main__.M::/torch.nn.modules.conv.Conv2d::conv # /media/pc/data/tmp/cache/conda/envs/py312x/lib/python3.12/site-packages/torch/nn/modules/conv.py:456:0\n",
      "  %/resize_2/Conv_output_0 : Float(1, 64, 24, 40, strides=[61440, 960, 40, 1], requires_grad=0, device=cpu) = onnx::Conv[dilations=[1, 1], group=1, kernel_shape=[1, 1], pads=[0, 0, 0, 0], strides=[2, 2], onnx_name=\"/resize_2/Conv\"](%/conv/Conv_output_0, %resize_2.weight), scope: __main__.M::/torch.nn.modules.conv.Conv2d::resize_2 # /media/pc/data/tmp/cache/conda/envs/py312x/lib/python3.12/site-packages/torch/nn/modules/conv.py:456:0\n",
      "  %/resize_4/Conv_output_0 : Float(1, 64, 12, 20, strides=[15360, 240, 20, 1], requires_grad=0, device=cpu) = onnx::Conv[dilations=[1, 1], group=1, kernel_shape=[1, 1], pads=[0, 0, 0, 0], strides=[4, 4], onnx_name=\"/resize_4/Conv\"](%/conv/Conv_output_0, %resize_4.weight), scope: __main__.M::/torch.nn.modules.conv.Conv2d::resize_4 # /media/pc/data/tmp/cache/conda/envs/py312x/lib/python3.12/site-packages/torch/nn/modules/conv.py:456:0\n",
      "  %/conv0/Conv_output_0 : Float(1, 36, 48, 80, strides=[138240, 3840, 80, 1], requires_grad=0, device=cpu) = onnx::Conv[dilations=[1, 1], group=1, kernel_shape=[1, 1], pads=[0, 0, 0, 0], strides=[1, 1], onnx_name=\"/conv0/Conv\"](%/conv/Conv_output_0, %conv0.weight), scope: __main__.M::/torch.nn.modules.conv.Conv2d::conv0 # /media/pc/data/tmp/cache/conda/envs/py312x/lib/python3.12/site-packages/torch/nn/modules/conv.py:456:0\n",
      "  %/Transpose_output_0 : Float(1, 48, 80, 36, strides=[138240, 80, 1, 3840], requires_grad=1, device=cpu) = onnx::Transpose[perm=[0, 2, 3, 1], onnx_name=\"/Transpose\"](%/conv0/Conv_output_0), scope: __main__.M:: # /tmp/ipykernel_1148509/1329727449.py:21:0\n",
      "  %/conv1/Conv_output_0 : Float(1, 36, 24, 40, strides=[34560, 960, 40, 1], requires_grad=0, device=cpu) = onnx::Conv[dilations=[1, 1], group=1, kernel_shape=[1, 1], pads=[0, 0, 0, 0], strides=[1, 1], onnx_name=\"/conv1/Conv\"](%/resize_2/Conv_output_0, %conv1.weight), scope: __main__.M::/torch.nn.modules.conv.Conv2d::conv1 # /media/pc/data/tmp/cache/conda/envs/py312x/lib/python3.12/site-packages/torch/nn/modules/conv.py:456:0\n",
      "  %/Transpose_1_output_0 : Float(1, 24, 40, 36, strides=[34560, 40, 1, 960], requires_grad=1, device=cpu) = onnx::Transpose[perm=[0, 2, 3, 1], onnx_name=\"/Transpose_1\"](%/conv1/Conv_output_0), scope: __main__.M:: # /tmp/ipykernel_1148509/1329727449.py:22:0\n",
      "  %/conv2/Conv_output_0 : Float(1, 48, 12, 20, strides=[11520, 240, 20, 1], requires_grad=0, device=cpu) = onnx::Conv[dilations=[1, 1], group=1, kernel_shape=[1, 1], pads=[0, 0, 0, 0], strides=[1, 1], onnx_name=\"/conv2/Conv\"](%/resize_4/Conv_output_0, %conv2.weight), scope: __main__.M::/torch.nn.modules.conv.Conv2d::conv2 # /media/pc/data/tmp/cache/conda/envs/py312x/lib/python3.12/site-packages/torch/nn/modules/conv.py:456:0\n",
      "  %/Transpose_2_output_0 : Float(1, 12, 20, 48, strides=[11520, 20, 1, 240], requires_grad=1, device=cpu) = onnx::Transpose[perm=[0, 2, 3, 1], onnx_name=\"/Transpose_2\"](%/conv2/Conv_output_0), scope: __main__.M:: # /tmp/ipykernel_1148509/1329727449.py:23:0\n",
      "  %/Constant_output_0 : Long(3, strides=[1], device=cpu) = onnx::Constant[value= 1 -1  3 [ CPULongType{3} ], onnx_name=\"/Constant\"](), scope: __main__.M:: # /tmp/ipykernel_1148509/1329727449.py:24:0\n",
      "  %/Reshape_output_0 : Float(1, 46080, 3, strides=[138240, 3, 1], requires_grad=1, device=cpu) = onnx::Reshape[allowzero=0, onnx_name=\"/Reshape\"](%/Transpose_output_0, %/Constant_output_0), scope: __main__.M:: # /tmp/ipykernel_1148509/1329727449.py:24:0\n",
      "  %/Constant_1_output_0 : Long(3, strides=[1], device=cpu) = onnx::Constant[value= 1 -1  3 [ CPULongType{3} ], onnx_name=\"/Constant_1\"](), scope: __main__.M:: # /tmp/ipykernel_1148509/1329727449.py:25:0\n",
      "  %/Reshape_1_output_0 : Float(1, 11520, 3, strides=[34560, 3, 1], requires_grad=1, device=cpu) = onnx::Reshape[allowzero=0, onnx_name=\"/Reshape_1\"](%/Transpose_1_output_0, %/Constant_1_output_0), scope: __main__.M:: # /tmp/ipykernel_1148509/1329727449.py:25:0\n",
      "  %/Constant_2_output_0 : Long(3, strides=[1], device=cpu) = onnx::Constant[value= 1 -1  3 [ CPULongType{3} ], onnx_name=\"/Constant_2\"](), scope: __main__.M:: # /tmp/ipykernel_1148509/1329727449.py:26:0\n",
      "  %/Reshape_2_output_0 : Float(1, 3840, 3, strides=[11520, 3, 1], requires_grad=1, device=cpu) = onnx::Reshape[allowzero=0, onnx_name=\"/Reshape_2\"](%/Transpose_2_output_0, %/Constant_2_output_0), scope: __main__.M:: # /tmp/ipykernel_1148509/1329727449.py:26:0\n",
      "  %/Concat_output_0 : Float(1, 61440, 3, strides=[184320, 3, 1], requires_grad=1, device=cpu) = onnx::Concat[axis=1, onnx_name=\"/Concat\"](%/Reshape_output_0, %/Reshape_1_output_0, %/Reshape_2_output_0), scope: __main__.M:: # /tmp/ipykernel_1148509/1329727449.py:27:0\n",
      "  %/conv00/Conv_output_0 : Float(1, 36, 48, 80, strides=[138240, 3840, 80, 1], requires_grad=0, device=cpu) = onnx::Conv[dilations=[1, 1], group=1, kernel_shape=[1, 1], pads=[0, 0, 0, 0], strides=[1, 1], onnx_name=\"/conv00/Conv\"](%/conv/Conv_output_0, %conv00.weight), scope: __main__.M::/torch.nn.modules.conv.Conv2d::conv00 # /media/pc/data/tmp/cache/conda/envs/py312x/lib/python3.12/site-packages/torch/nn/modules/conv.py:456:0\n",
      "  %/Transpose_3_output_0 : Float(1, 48, 80, 36, strides=[138240, 80, 1, 3840], requires_grad=1, device=cpu) = onnx::Transpose[perm=[0, 2, 3, 1], onnx_name=\"/Transpose_3\"](%/conv00/Conv_output_0), scope: __main__.M:: # /tmp/ipykernel_1148509/1329727449.py:29:0\n",
      "  %/conv11/Conv_output_0 : Float(1, 36, 24, 40, strides=[34560, 960, 40, 1], requires_grad=0, device=cpu) = onnx::Conv[dilations=[1, 1], group=1, kernel_shape=[1, 1], pads=[0, 0, 0, 0], strides=[1, 1], onnx_name=\"/conv11/Conv\"](%/resize_2/Conv_output_0, %conv11.weight), scope: __main__.M::/torch.nn.modules.conv.Conv2d::conv11 # /media/pc/data/tmp/cache/conda/envs/py312x/lib/python3.12/site-packages/torch/nn/modules/conv.py:456:0\n",
      "  %/Transpose_4_output_0 : Float(1, 24, 40, 36, strides=[34560, 40, 1, 960], requires_grad=1, device=cpu) = onnx::Transpose[perm=[0, 2, 3, 1], onnx_name=\"/Transpose_4\"](%/conv11/Conv_output_0), scope: __main__.M:: # /tmp/ipykernel_1148509/1329727449.py:30:0\n",
      "  %/conv22/Conv_output_0 : Float(1, 48, 12, 20, strides=[11520, 240, 20, 1], requires_grad=0, device=cpu) = onnx::Conv[dilations=[1, 1], group=1, kernel_shape=[1, 1], pads=[0, 0, 0, 0], strides=[1, 1], onnx_name=\"/conv22/Conv\"](%/resize_4/Conv_output_0, %conv22.weight), scope: __main__.M::/torch.nn.modules.conv.Conv2d::conv22 # /media/pc/data/tmp/cache/conda/envs/py312x/lib/python3.12/site-packages/torch/nn/modules/conv.py:456:0\n",
      "  %/Transpose_5_output_0 : Float(1, 12, 20, 48, strides=[11520, 20, 1, 240], requires_grad=1, device=cpu) = onnx::Transpose[perm=[0, 2, 3, 1], onnx_name=\"/Transpose_5\"](%/conv22/Conv_output_0), scope: __main__.M:: # /tmp/ipykernel_1148509/1329727449.py:31:0\n",
      "  %/Constant_3_output_0 : Long(3, strides=[1], device=cpu) = onnx::Constant[value= 1 -1  4 [ CPULongType{3} ], onnx_name=\"/Constant_3\"](), scope: __main__.M:: # /tmp/ipykernel_1148509/1329727449.py:32:0\n",
      "  %/Reshape_3_output_0 : Float(1, 34560, 4, strides=[138240, 4, 1], requires_grad=1, device=cpu) = onnx::Reshape[allowzero=0, onnx_name=\"/Reshape_3\"](%/Transpose_3_output_0, %/Constant_3_output_0), scope: __main__.M:: # /tmp/ipykernel_1148509/1329727449.py:32:0\n",
      "  %/Constant_4_output_0 : Long(3, strides=[1], device=cpu) = onnx::Constant[value= 1 -1  4 [ CPULongType{3} ], onnx_name=\"/Constant_4\"](), scope: __main__.M:: # /tmp/ipykernel_1148509/1329727449.py:33:0\n",
      "  %/Reshape_4_output_0 : Float(1, 8640, 4, strides=[34560, 4, 1], requires_grad=1, device=cpu) = onnx::Reshape[allowzero=0, onnx_name=\"/Reshape_4\"](%/Transpose_4_output_0, %/Constant_4_output_0), scope: __main__.M:: # /tmp/ipykernel_1148509/1329727449.py:33:0\n",
      "  %/Constant_5_output_0 : Long(3, strides=[1], device=cpu) = onnx::Constant[value= 1 -1  4 [ CPULongType{3} ], onnx_name=\"/Constant_5\"](), scope: __main__.M:: # /tmp/ipykernel_1148509/1329727449.py:34:0\n",
      "  %/Reshape_5_output_0 : Float(1, 2880, 4, strides=[11520, 4, 1], requires_grad=1, device=cpu) = onnx::Reshape[allowzero=0, onnx_name=\"/Reshape_5\"](%/Transpose_5_output_0, %/Constant_5_output_0), scope: __main__.M:: # /tmp/ipykernel_1148509/1329727449.py:34:0\n",
      "  %38 : Float(1, 46080, 4, strides=[184320, 4, 1], requires_grad=1, device=cpu) = onnx::Concat[axis=1, onnx_name=\"/Concat_1\"](%/Reshape_3_output_0, %/Reshape_4_output_0, %/Reshape_5_output_0), scope: __main__.M:: # /tmp/ipykernel_1148509/1329727449.py:35:0\n",
      "  %output : Float(1, 61440, 3, strides=[184320, 3, 1], requires_grad=1, device=cpu) = onnx::Softmax[axis=2, onnx_name=\"/Softmax\"](%/Concat_output_0), scope: __main__.M:: # /tmp/ipykernel_1148509/1329727449.py:36:0\n",
      "  return (%output, %38)\n",
      "\n"
     ]
    }
   ],
   "source": [
    "from torch.onnx import OperatorExportTypes, utils\n",
    "import numpy as np\n",
    "import onnx\n",
    "\n",
    "model = M()\n",
    "model.eval()\n",
    "\n",
    "config = Config((1, 3, 48, 80))\n",
    "data_np = np.random.rand(*config.shape).astype(config.dtype)\n",
    "output_name = \"test-det-class\"\n",
    "xx = torch.rand(*config.shape, dtype=torch.float32, requires_grad=False)\n",
    "utils.export(\n",
    "    model,               # torch 模型\n",
    "    xx,                         # 模型输入或者对于多个输入，使用元组\n",
    "    f\"{root_dir}/{output_name}.onnx\",               # 模型保存的位置（可以是文件或类似文件的对象）\n",
    "    export_params=True,        # 将训练后的参数权重存储在模型文件内\n",
    "    # opset_version=17,          # 导出模型的 ONNX 版本\n",
    "    do_constant_folding=True,  # 是否执行常量折叠以进行优化\n",
    "    input_names = [config.name],    # 模型的输入名称\n",
    "    output_names = ['output'], # 模型的输出名称\n",
    "    # keep_initializers_as_inputs=True,\n",
    "    # export_modules_as_functions=True,\n",
    "    verbose=True,\n",
    "    # operator_export_type=OperatorExportTypes.ONNX_FALLTHROUGH,\n",
    "    # dynamic_axes={'data' : {0 : 'batch_size'},    # 可变长度的轴\n",
    "    #               'output' : {0 : 'batch_size'}}\n",
    ")"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## 前端导入"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "ONNX 前端："
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 5,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/html": [
       "<div class=\"highlight\" style=\"background: \"><pre style=\"line-height: 125%;\"><span></span><span style=\"color: #008000; font-weight: bold\">def</span> <span style=\"color: #AA22FF\">@main</span>(<span style=\"color: #AA22FF; font-weight: bold\">%</span>data: Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">3</span>, <span style=\"color: #008000\">48</span>, <span style=\"color: #008000\">80</span>), float32] <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">3</span>, <span style=\"color: #008000\">48</span>, <span style=\"color: #008000\">80</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>conv<span style=\"color: #AA22FF; font-weight: bold\">/</span>Conv<span style=\"color: #AA22FF; font-weight: bold\">.</span>data:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>) <span style=\"color: #AA22FF; font-weight: bold\">-&gt;</span> (Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">61440</span>, <span style=\"color: #008000\">3</span>), float32], Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">46080</span>, <span style=\"color: #008000\">4</span>), float32]) {\n",
       "  <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">=</span> nn<span style=\"color: #AA22FF; font-weight: bold\">.</span>conv2d(<span style=\"color: #AA22FF; font-weight: bold\">%</span>data, meta[relay<span style=\"color: #AA22FF; font-weight: bold\">.</span>Constant][<span style=\"color: #008000\">0</span>] <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">64</span>, <span style=\"color: #008000\">3</span>, <span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>conv<span style=\"color: #AA22FF; font-weight: bold\">/</span>Conv<span style=\"color: #AA22FF; font-weight: bold\">.</span>conv<span style=\"color: #AA22FF; font-weight: bold\">.</span>weight:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>, padding<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>], channels<span style=\"color: #AA22FF; font-weight: bold\">=</span><span style=\"color: #008000\">64</span>, kernel_size<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1</span>]) <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">64</span>, <span style=\"color: #008000\">48</span>, <span style=\"color: #008000\">80</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>conv<span style=\"color: #AA22FF; font-weight: bold\">/</span>Conv:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>;\n",
       "  <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">1</span> <span style=\"color: #AA22FF; font-weight: bold\">=</span> nn<span style=\"color: #AA22FF; font-weight: bold\">.</span>conv2d(<span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">0</span>, meta[relay<span style=\"color: #AA22FF; font-weight: bold\">.</span>Constant][<span style=\"color: #008000\">1</span>] <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">36</span>, <span style=\"color: #008000\">64</span>, <span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>conv0<span style=\"color: #AA22FF; font-weight: bold\">/</span>Conv<span style=\"color: #AA22FF; font-weight: bold\">.</span>conv0<span style=\"color: #AA22FF; font-weight: bold\">.</span>weight:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>, padding<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>], channels<span style=\"color: #AA22FF; font-weight: bold\">=</span><span style=\"color: #008000\">36</span>, kernel_size<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1</span>]) <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">36</span>, <span style=\"color: #008000\">48</span>, <span style=\"color: #008000\">80</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>conv0<span style=\"color: #AA22FF; font-weight: bold\">/</span>Conv:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>;\n",
       "  <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">2</span> <span style=\"color: #AA22FF; font-weight: bold\">=</span> transpose(<span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">1</span>, axes<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">2</span>, <span style=\"color: #008000\">3</span>, <span style=\"color: #008000\">1</span>]) <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">48</span>, <span style=\"color: #008000\">80</span>, <span style=\"color: #008000\">36</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>Transpose:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>;\n",
       "  <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">3</span> <span style=\"color: #AA22FF; font-weight: bold\">=</span> nn<span style=\"color: #AA22FF; font-weight: bold\">.</span>conv2d(<span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">0</span>, meta[relay<span style=\"color: #AA22FF; font-weight: bold\">.</span>Constant][<span style=\"color: #008000\">2</span>] <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">64</span>, <span style=\"color: #008000\">64</span>, <span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>resize_2<span style=\"color: #AA22FF; font-weight: bold\">/</span>Conv<span style=\"color: #AA22FF; font-weight: bold\">.</span>resize_2<span style=\"color: #AA22FF; font-weight: bold\">.</span>weight:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>, strides<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">2</span>, <span style=\"color: #008000\">2</span>], padding<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>], channels<span style=\"color: #AA22FF; font-weight: bold\">=</span><span style=\"color: #008000\">64</span>, kernel_size<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1</span>]) <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">64</span>, <span style=\"color: #008000\">24</span>, <span style=\"color: #008000\">40</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>resize_2<span style=\"color: #AA22FF; font-weight: bold\">/</span>Conv:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>;\n",
       "  <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">4</span> <span style=\"color: #AA22FF; font-weight: bold\">=</span> nn<span style=\"color: #AA22FF; font-weight: bold\">.</span>conv2d(<span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">3</span>, meta[relay<span style=\"color: #AA22FF; font-weight: bold\">.</span>Constant][<span style=\"color: #008000\">3</span>] <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">36</span>, <span style=\"color: #008000\">64</span>, <span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>conv1<span style=\"color: #AA22FF; font-weight: bold\">/</span>Conv<span style=\"color: #AA22FF; font-weight: bold\">.</span>conv1<span style=\"color: #AA22FF; font-weight: bold\">.</span>weight:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>, padding<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>], channels<span style=\"color: #AA22FF; font-weight: bold\">=</span><span style=\"color: #008000\">36</span>, kernel_size<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1</span>]) <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">36</span>, <span style=\"color: #008000\">24</span>, <span style=\"color: #008000\">40</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>conv1<span style=\"color: #AA22FF; font-weight: bold\">/</span>Conv:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>;\n",
       "  <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">5</span> <span style=\"color: #AA22FF; font-weight: bold\">=</span> transpose(<span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">4</span>, axes<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">2</span>, <span style=\"color: #008000\">3</span>, <span style=\"color: #008000\">1</span>]) <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">24</span>, <span style=\"color: #008000\">40</span>, <span style=\"color: #008000\">36</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>Transpose_1:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>;\n",
       "  <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">6</span> <span style=\"color: #AA22FF; font-weight: bold\">=</span> nn<span style=\"color: #AA22FF; font-weight: bold\">.</span>conv2d(<span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">0</span>, meta[relay<span style=\"color: #AA22FF; font-weight: bold\">.</span>Constant][<span style=\"color: #008000\">4</span>] <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">64</span>, <span style=\"color: #008000\">64</span>, <span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>resize_4<span style=\"color: #AA22FF; font-weight: bold\">/</span>Conv<span style=\"color: #AA22FF; font-weight: bold\">.</span>resize_4<span style=\"color: #AA22FF; font-weight: bold\">.</span>weight:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>, strides<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">4</span>, <span style=\"color: #008000\">4</span>], padding<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>], channels<span style=\"color: #AA22FF; font-weight: bold\">=</span><span style=\"color: #008000\">64</span>, kernel_size<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1</span>]) <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">64</span>, <span style=\"color: #008000\">12</span>, <span style=\"color: #008000\">20</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>resize_4<span style=\"color: #AA22FF; font-weight: bold\">/</span>Conv:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>;\n",
       "  <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">7</span> <span style=\"color: #AA22FF; font-weight: bold\">=</span> nn<span style=\"color: #AA22FF; font-weight: bold\">.</span>conv2d(<span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">6</span>, meta[relay<span style=\"color: #AA22FF; font-weight: bold\">.</span>Constant][<span style=\"color: #008000\">5</span>] <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">48</span>, <span style=\"color: #008000\">64</span>, <span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>conv2<span style=\"color: #AA22FF; font-weight: bold\">/</span>Conv<span style=\"color: #AA22FF; font-weight: bold\">.</span>conv2<span style=\"color: #AA22FF; font-weight: bold\">.</span>weight:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>, padding<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>], channels<span style=\"color: #AA22FF; font-weight: bold\">=</span><span style=\"color: #008000\">48</span>, kernel_size<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1</span>]) <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">48</span>, <span style=\"color: #008000\">12</span>, <span style=\"color: #008000\">20</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>conv2<span style=\"color: #AA22FF; font-weight: bold\">/</span>Conv:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>;\n",
       "  <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">8</span> <span style=\"color: #AA22FF; font-weight: bold\">=</span> transpose(<span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">7</span>, axes<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">2</span>, <span style=\"color: #008000\">3</span>, <span style=\"color: #008000\">1</span>]) <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">12</span>, <span style=\"color: #008000\">20</span>, <span style=\"color: #008000\">48</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>Transpose_2:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>;\n",
       "  <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">9</span> <span style=\"color: #AA22FF; font-weight: bold\">=</span> reshape(<span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">2</span>, newshape<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">1</span>, <span style=\"color: #AA22FF; font-weight: bold\">-</span><span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">3</span>]) <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">46080</span>, <span style=\"color: #008000\">3</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>Reshape:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>;\n",
       "  <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">10</span> <span style=\"color: #AA22FF; font-weight: bold\">=</span> reshape(<span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">5</span>, newshape<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">1</span>, <span style=\"color: #AA22FF; font-weight: bold\">-</span><span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">3</span>]) <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">11520</span>, <span style=\"color: #008000\">3</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>Reshape_1:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>;\n",
       "  <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">11</span> <span style=\"color: #AA22FF; font-weight: bold\">=</span> reshape(<span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">8</span>, newshape<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">1</span>, <span style=\"color: #AA22FF; font-weight: bold\">-</span><span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">3</span>]) <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">3840</span>, <span style=\"color: #008000\">3</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>Reshape_2:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>;\n",
       "  <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">12</span> <span style=\"color: #AA22FF; font-weight: bold\">=</span> (<span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">9</span>, <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">10</span>, <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">11</span>) <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>(Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">46080</span>, <span style=\"color: #008000\">3</span>), float32], Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">11520</span>, <span style=\"color: #008000\">3</span>), float32], Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">3840</span>, <span style=\"color: #008000\">3</span>), float32]) span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>Concat:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>;\n",
       "  <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">13</span> <span style=\"color: #AA22FF; font-weight: bold\">=</span> concatenate(<span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">12</span>, axis<span style=\"color: #AA22FF; font-weight: bold\">=</span><span style=\"color: #008000\">1</span>) <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">61440</span>, <span style=\"color: #008000\">3</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>Concat:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>;\n",
       "  <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">14</span> <span style=\"color: #AA22FF; font-weight: bold\">=</span> nn<span style=\"color: #AA22FF; font-weight: bold\">.</span>conv2d(<span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">0</span>, meta[relay<span style=\"color: #AA22FF; font-weight: bold\">.</span>Constant][<span style=\"color: #008000\">6</span>] <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">36</span>, <span style=\"color: #008000\">64</span>, <span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>conv00<span style=\"color: #AA22FF; font-weight: bold\">/</span>Conv<span style=\"color: #AA22FF; font-weight: bold\">.</span>conv00<span style=\"color: #AA22FF; font-weight: bold\">.</span>weight:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>, padding<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>], channels<span style=\"color: #AA22FF; font-weight: bold\">=</span><span style=\"color: #008000\">36</span>, kernel_size<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1</span>]) <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">36</span>, <span style=\"color: #008000\">48</span>, <span style=\"color: #008000\">80</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>conv00<span style=\"color: #AA22FF; font-weight: bold\">/</span>Conv:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>;\n",
       "  <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">15</span> <span style=\"color: #AA22FF; font-weight: bold\">=</span> transpose(<span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">14</span>, axes<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">2</span>, <span style=\"color: #008000\">3</span>, <span style=\"color: #008000\">1</span>]) <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">48</span>, <span style=\"color: #008000\">80</span>, <span style=\"color: #008000\">36</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>Transpose_3:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>;\n",
       "  <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">16</span> <span style=\"color: #AA22FF; font-weight: bold\">=</span> nn<span style=\"color: #AA22FF; font-weight: bold\">.</span>conv2d(<span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">3</span>, meta[relay<span style=\"color: #AA22FF; font-weight: bold\">.</span>Constant][<span style=\"color: #008000\">7</span>] <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">36</span>, <span style=\"color: #008000\">64</span>, <span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>conv11<span style=\"color: #AA22FF; font-weight: bold\">/</span>Conv<span style=\"color: #AA22FF; font-weight: bold\">.</span>conv11<span style=\"color: #AA22FF; font-weight: bold\">.</span>weight:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>, padding<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>], channels<span style=\"color: #AA22FF; font-weight: bold\">=</span><span style=\"color: #008000\">36</span>, kernel_size<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1</span>]) <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">36</span>, <span style=\"color: #008000\">24</span>, <span style=\"color: #008000\">40</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>conv11<span style=\"color: #AA22FF; font-weight: bold\">/</span>Conv:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>;\n",
       "  <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">17</span> <span style=\"color: #AA22FF; font-weight: bold\">=</span> transpose(<span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">16</span>, axes<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">2</span>, <span style=\"color: #008000\">3</span>, <span style=\"color: #008000\">1</span>]) <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">24</span>, <span style=\"color: #008000\">40</span>, <span style=\"color: #008000\">36</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>Transpose_4:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>;\n",
       "  <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">18</span> <span style=\"color: #AA22FF; font-weight: bold\">=</span> nn<span style=\"color: #AA22FF; font-weight: bold\">.</span>conv2d(<span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">6</span>, meta[relay<span style=\"color: #AA22FF; font-weight: bold\">.</span>Constant][<span style=\"color: #008000\">8</span>] <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">48</span>, <span style=\"color: #008000\">64</span>, <span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>conv22<span style=\"color: #AA22FF; font-weight: bold\">/</span>Conv<span style=\"color: #AA22FF; font-weight: bold\">.</span>conv22<span style=\"color: #AA22FF; font-weight: bold\">.</span>weight:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>, padding<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>], channels<span style=\"color: #AA22FF; font-weight: bold\">=</span><span style=\"color: #008000\">48</span>, kernel_size<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1</span>]) <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">48</span>, <span style=\"color: #008000\">12</span>, <span style=\"color: #008000\">20</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>conv22<span style=\"color: #AA22FF; font-weight: bold\">/</span>Conv:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>;\n",
       "  <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">19</span> <span style=\"color: #AA22FF; font-weight: bold\">=</span> transpose(<span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">18</span>, axes<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">2</span>, <span style=\"color: #008000\">3</span>, <span style=\"color: #008000\">1</span>]) <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">12</span>, <span style=\"color: #008000\">20</span>, <span style=\"color: #008000\">48</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>Transpose_5:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>;\n",
       "  <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">20</span> <span style=\"color: #AA22FF; font-weight: bold\">=</span> reshape(<span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">15</span>, newshape<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">1</span>, <span style=\"color: #AA22FF; font-weight: bold\">-</span><span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">4</span>]) <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">34560</span>, <span style=\"color: #008000\">4</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>Reshape_3:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>;\n",
       "  <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">21</span> <span style=\"color: #AA22FF; font-weight: bold\">=</span> reshape(<span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">17</span>, newshape<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">1</span>, <span style=\"color: #AA22FF; font-weight: bold\">-</span><span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">4</span>]) <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">8640</span>, <span style=\"color: #008000\">4</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>Reshape_4:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>;\n",
       "  <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">22</span> <span style=\"color: #AA22FF; font-weight: bold\">=</span> reshape(<span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">19</span>, newshape<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">1</span>, <span style=\"color: #AA22FF; font-weight: bold\">-</span><span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">4</span>]) <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">2880</span>, <span style=\"color: #008000\">4</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>Reshape_5:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>;\n",
       "  <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">23</span> <span style=\"color: #AA22FF; font-weight: bold\">=</span> (<span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">20</span>, <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">21</span>, <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">22</span>) <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>(Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">34560</span>, <span style=\"color: #008000\">4</span>), float32], Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">8640</span>, <span style=\"color: #008000\">4</span>), float32], Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">2880</span>, <span style=\"color: #008000\">4</span>), float32]) span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>Concat_1:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>;\n",
       "  <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">24</span> <span style=\"color: #AA22FF; font-weight: bold\">=</span> nn<span style=\"color: #AA22FF; font-weight: bold\">.</span>softmax(<span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">13</span>, axis<span style=\"color: #AA22FF; font-weight: bold\">=</span><span style=\"color: #008000\">2</span>) <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">61440</span>, <span style=\"color: #008000\">3</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>Softmax:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>;\n",
       "  <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">25</span> <span style=\"color: #AA22FF; font-weight: bold\">=</span> concatenate(<span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">23</span>, axis<span style=\"color: #AA22FF; font-weight: bold\">=</span><span style=\"color: #008000\">1</span>) <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">46080</span>, <span style=\"color: #008000\">4</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>Concat_1:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>;\n",
       "  (<span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">24</span>, <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">25</span>) <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>(Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">61440</span>, <span style=\"color: #008000\">3</span>), float32], Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">46080</span>, <span style=\"color: #008000\">4</span>), float32]) <span style=\"color: #AA22FF; font-weight: bold\">*/</span>\n",
       "}\n",
       "</pre></div>\n"
      ],
      "text/plain": [
       "<IPython.core.display.HTML object>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "from tvm import relay\n",
    "onnx_model = onnx.load(f\"{root_dir}/{output_name}.onnx\")\n",
    "mod, params = relay.frontend.from_onnx(onnx_model, {config.name: config.shape}, freeze_params=True)\n",
    "onnx_mod = relay.transform.InferType()(mod)\n",
    "onnx_mod.show()"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "验证数值一致性："
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 6,
   "metadata": {},
   "outputs": [],
   "source": [
    "import onnxruntime\n",
    "\n",
    "ort_session = onnxruntime.InferenceSession(f\"{root_dir}/{output_name}.onnx\", providers=[\"CPUExecutionProvider\"])\n",
    "# compute ONNX Runtime output prediction\n",
    "ort_inputs = {ort_session.get_inputs()[0].name: data_np}\n",
    "ort_outputs = ort_session.run(None, ort_inputs)\n",
    "with torch.no_grad():\n",
    "    torch_outputs = model(torch.from_numpy(data_np))\n",
    "    torch_outputs = [o.numpy() for o in torch_outputs]\n",
    "[\n",
    "    np.testing.assert_allclose(ort_output, torch_output, rtol=1e-07, atol=1e-5) \n",
    "    for ort_output, torch_output in zip(ort_outputs, torch_outputs)\n",
    "];"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 7,
   "metadata": {},
   "outputs": [],
   "source": [
    "import logging\n",
    "import textwrap\n",
    "import numpy as np\n",
    "from tvm.relay.testing import run_infer_type\n",
    "from tvm.relay.dataflow_pattern import (\n",
    "    wildcard, is_op, is_tuple,\n",
    "    is_constant, is_tuple_get_item,\n",
    "    DFPatternCallback,\n",
    "    rewrite\n",
    ")\n",
    "import tvm\n",
    "from tvm.relay import transform as _transform"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 8,
   "metadata": {},
   "outputs": [],
   "source": [
    "from special_op import det_class_predict, det_bbox_predict"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 9,
   "metadata": {},
   "outputs": [],
   "source": [
    "class TransposeReshapeConcatRewrite(DFPatternCallback):\n",
    "    \"\"\"融合 Transpose+Reshape+Concat+Softmax => det_class_predict\n",
    "    \"\"\"\n",
    "    def __init__(self):\n",
    "        super().__init__()\n",
    "        axes = (0, 2, 3, 1)\n",
    "        # newshape = (1, -1, group)\n",
    "        self.x0 = wildcard()\n",
    "        self.x1 = wildcard()\n",
    "        self.x2 = wildcard()\n",
    "        self.transpose0 = is_op(\"transpose\")(self.x0).has_attr({\"axes\": axes})\n",
    "        self.reshape0 = is_op(\"reshape\")(self.transpose0) #.has_attr({\"newshape\": newshape})\n",
    "        self.transpose1 = is_op(\"transpose\")(self.x1).has_attr({\"axes\": axes})\n",
    "        self.reshape1 = is_op(\"reshape\")(self.transpose1) #.has_attr({\"newshape\": newshape})\n",
    "        self.transpose2 = is_op(\"transpose\")(self.x2).has_attr({\"axes\": axes})\n",
    "        self.reshape2 = is_op(\"reshape\")(self.transpose2) #.has_attr({\"newshape\": newshape})\n",
    "        self.tuple_op = is_tuple((self.reshape0, self.reshape1, self.reshape2))\n",
    "        self.cat = is_op(\"concatenate\")(self.tuple_op).has_attr({\"axis\": 1})\n",
    "        self.softmax = is_op(\"nn.softmax\")(self.cat)\n",
    "        self.output = self.softmax | self.cat\n",
    "        self.pattern = self.output\n",
    "\n",
    "    def callback(self, pre, post, node_map):\n",
    "        x0 = node_map[self.x0][0]\n",
    "        x1 = node_map[self.x1][0]\n",
    "        x2 = node_map[self.x2][0]\n",
    "        reshape2 = node_map[self.reshape2][0]\n",
    "        shape = _transform.InferTypeLocal(reshape2).shape\n",
    "        softmax = node_map.get(self.softmax, [])\n",
    "        if softmax:\n",
    "            return det_class_predict(x0, x1, x2, class_num=int(shape[-1]))\n",
    "        else:\n",
    "            return det_bbox_predict(x0, x1, x2)\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 10,
   "metadata": {},
   "outputs": [],
   "source": [
    "from copy import deepcopy\n",
    "\n",
    "origin_mod = deepcopy(mod)\n",
    "mod[\"main\"] = rewrite(TransposeReshapeConcatRewrite(), mod[\"main\"])"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 11,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/html": [
       "<div class=\"highlight\" style=\"background: \"><pre style=\"line-height: 125%;\"><span></span><span style=\"color: #008000; font-weight: bold\">def</span> <span style=\"color: #AA22FF\">@main</span>(<span style=\"color: #AA22FF; font-weight: bold\">%</span>data: Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">3</span>, <span style=\"color: #008000\">48</span>, <span style=\"color: #008000\">80</span>), float32] <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">3</span>, <span style=\"color: #008000\">48</span>, <span style=\"color: #008000\">80</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>conv<span style=\"color: #AA22FF; font-weight: bold\">/</span>Conv<span style=\"color: #AA22FF; font-weight: bold\">.</span>data:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>) <span style=\"color: #AA22FF; font-weight: bold\">-&gt;</span> (Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">61440</span>, <span style=\"color: #008000\">3</span>), float32], Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">46080</span>, <span style=\"color: #008000\">4</span>), float32]) {\n",
       "  <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">=</span> nn<span style=\"color: #AA22FF; font-weight: bold\">.</span>conv2d(<span style=\"color: #AA22FF; font-weight: bold\">%</span>data, meta[relay<span style=\"color: #AA22FF; font-weight: bold\">.</span>Constant][<span style=\"color: #008000\">0</span>] <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">64</span>, <span style=\"color: #008000\">3</span>, <span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>conv<span style=\"color: #AA22FF; font-weight: bold\">/</span>Conv<span style=\"color: #AA22FF; font-weight: bold\">.</span>conv<span style=\"color: #AA22FF; font-weight: bold\">.</span>weight:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>, padding<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>], channels<span style=\"color: #AA22FF; font-weight: bold\">=</span><span style=\"color: #008000\">64</span>, kernel_size<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1</span>]) <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">64</span>, <span style=\"color: #008000\">48</span>, <span style=\"color: #008000\">80</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>conv<span style=\"color: #AA22FF; font-weight: bold\">/</span>Conv:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>;\n",
       "  <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">1</span> <span style=\"color: #AA22FF; font-weight: bold\">=</span> nn<span style=\"color: #AA22FF; font-weight: bold\">.</span>conv2d(<span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">0</span>, meta[relay<span style=\"color: #AA22FF; font-weight: bold\">.</span>Constant][<span style=\"color: #008000\">2</span>] <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">64</span>, <span style=\"color: #008000\">64</span>, <span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>resize_2<span style=\"color: #AA22FF; font-weight: bold\">/</span>Conv<span style=\"color: #AA22FF; font-weight: bold\">.</span>resize_2<span style=\"color: #AA22FF; font-weight: bold\">.</span>weight:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>, strides<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">2</span>, <span style=\"color: #008000\">2</span>], padding<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>], channels<span style=\"color: #AA22FF; font-weight: bold\">=</span><span style=\"color: #008000\">64</span>, kernel_size<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1</span>]) <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">64</span>, <span style=\"color: #008000\">24</span>, <span style=\"color: #008000\">40</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>resize_2<span style=\"color: #AA22FF; font-weight: bold\">/</span>Conv:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>;\n",
       "  <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">2</span> <span style=\"color: #AA22FF; font-weight: bold\">=</span> nn<span style=\"color: #AA22FF; font-weight: bold\">.</span>conv2d(<span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">0</span>, meta[relay<span style=\"color: #AA22FF; font-weight: bold\">.</span>Constant][<span style=\"color: #008000\">4</span>] <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">64</span>, <span style=\"color: #008000\">64</span>, <span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>resize_4<span style=\"color: #AA22FF; font-weight: bold\">/</span>Conv<span style=\"color: #AA22FF; font-weight: bold\">.</span>resize_4<span style=\"color: #AA22FF; font-weight: bold\">.</span>weight:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>, strides<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">4</span>, <span style=\"color: #008000\">4</span>], padding<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>], channels<span style=\"color: #AA22FF; font-weight: bold\">=</span><span style=\"color: #008000\">64</span>, kernel_size<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1</span>]) <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">64</span>, <span style=\"color: #008000\">12</span>, <span style=\"color: #008000\">20</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>resize_4<span style=\"color: #AA22FF; font-weight: bold\">/</span>Conv:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>;\n",
       "  <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">3</span> <span style=\"color: #AA22FF; font-weight: bold\">=</span> nn<span style=\"color: #AA22FF; font-weight: bold\">.</span>conv2d(<span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">0</span>, meta[relay<span style=\"color: #AA22FF; font-weight: bold\">.</span>Constant][<span style=\"color: #008000\">1</span>] <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">36</span>, <span style=\"color: #008000\">64</span>, <span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>conv0<span style=\"color: #AA22FF; font-weight: bold\">/</span>Conv<span style=\"color: #AA22FF; font-weight: bold\">.</span>conv0<span style=\"color: #AA22FF; font-weight: bold\">.</span>weight:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>, padding<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>], channels<span style=\"color: #AA22FF; font-weight: bold\">=</span><span style=\"color: #008000\">36</span>, kernel_size<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1</span>]) <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">36</span>, <span style=\"color: #008000\">48</span>, <span style=\"color: #008000\">80</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>conv0<span style=\"color: #AA22FF; font-weight: bold\">/</span>Conv:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>;\n",
       "  <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">4</span> <span style=\"color: #AA22FF; font-weight: bold\">=</span> nn<span style=\"color: #AA22FF; font-weight: bold\">.</span>conv2d(<span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">1</span>, meta[relay<span style=\"color: #AA22FF; font-weight: bold\">.</span>Constant][<span style=\"color: #008000\">3</span>] <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">36</span>, <span style=\"color: #008000\">64</span>, <span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>conv1<span style=\"color: #AA22FF; font-weight: bold\">/</span>Conv<span style=\"color: #AA22FF; font-weight: bold\">.</span>conv1<span style=\"color: #AA22FF; font-weight: bold\">.</span>weight:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>, padding<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>], channels<span style=\"color: #AA22FF; font-weight: bold\">=</span><span style=\"color: #008000\">36</span>, kernel_size<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1</span>]) <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">36</span>, <span style=\"color: #008000\">24</span>, <span style=\"color: #008000\">40</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>conv1<span style=\"color: #AA22FF; font-weight: bold\">/</span>Conv:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>;\n",
       "  <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">5</span> <span style=\"color: #AA22FF; font-weight: bold\">=</span> nn<span style=\"color: #AA22FF; font-weight: bold\">.</span>conv2d(<span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">2</span>, meta[relay<span style=\"color: #AA22FF; font-weight: bold\">.</span>Constant][<span style=\"color: #008000\">5</span>] <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">48</span>, <span style=\"color: #008000\">64</span>, <span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>conv2<span style=\"color: #AA22FF; font-weight: bold\">/</span>Conv<span style=\"color: #AA22FF; font-weight: bold\">.</span>conv2<span style=\"color: #AA22FF; font-weight: bold\">.</span>weight:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>, padding<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>], channels<span style=\"color: #AA22FF; font-weight: bold\">=</span><span style=\"color: #008000\">48</span>, kernel_size<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1</span>]) <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">48</span>, <span style=\"color: #008000\">12</span>, <span style=\"color: #008000\">20</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>conv2<span style=\"color: #AA22FF; font-weight: bold\">/</span>Conv:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>;\n",
       "  <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">6</span> <span style=\"color: #AA22FF; font-weight: bold\">=</span> nn<span style=\"color: #AA22FF; font-weight: bold\">.</span>conv2d(<span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">0</span>, meta[relay<span style=\"color: #AA22FF; font-weight: bold\">.</span>Constant][<span style=\"color: #008000\">6</span>] <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">36</span>, <span style=\"color: #008000\">64</span>, <span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>conv00<span style=\"color: #AA22FF; font-weight: bold\">/</span>Conv<span style=\"color: #AA22FF; font-weight: bold\">.</span>conv00<span style=\"color: #AA22FF; font-weight: bold\">.</span>weight:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>, padding<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>], channels<span style=\"color: #AA22FF; font-weight: bold\">=</span><span style=\"color: #008000\">36</span>, kernel_size<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1</span>]) <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">36</span>, <span style=\"color: #008000\">48</span>, <span style=\"color: #008000\">80</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>conv00<span style=\"color: #AA22FF; font-weight: bold\">/</span>Conv:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>;\n",
       "  <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">7</span> <span style=\"color: #AA22FF; font-weight: bold\">=</span> nn<span style=\"color: #AA22FF; font-weight: bold\">.</span>conv2d(<span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">1</span>, meta[relay<span style=\"color: #AA22FF; font-weight: bold\">.</span>Constant][<span style=\"color: #008000\">7</span>] <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">36</span>, <span style=\"color: #008000\">64</span>, <span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>conv11<span style=\"color: #AA22FF; font-weight: bold\">/</span>Conv<span style=\"color: #AA22FF; font-weight: bold\">.</span>conv11<span style=\"color: #AA22FF; font-weight: bold\">.</span>weight:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>, padding<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>], channels<span style=\"color: #AA22FF; font-weight: bold\">=</span><span style=\"color: #008000\">36</span>, kernel_size<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1</span>]) <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">36</span>, <span style=\"color: #008000\">24</span>, <span style=\"color: #008000\">40</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>conv11<span style=\"color: #AA22FF; font-weight: bold\">/</span>Conv:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>;\n",
       "  <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">8</span> <span style=\"color: #AA22FF; font-weight: bold\">=</span> nn<span style=\"color: #AA22FF; font-weight: bold\">.</span>conv2d(<span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">2</span>, meta[relay<span style=\"color: #AA22FF; font-weight: bold\">.</span>Constant][<span style=\"color: #008000\">8</span>] <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">48</span>, <span style=\"color: #008000\">64</span>, <span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>conv22<span style=\"color: #AA22FF; font-weight: bold\">/</span>Conv<span style=\"color: #AA22FF; font-weight: bold\">.</span>conv22<span style=\"color: #AA22FF; font-weight: bold\">.</span>weight:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>, padding<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>, <span style=\"color: #008000\">0</span>], channels<span style=\"color: #AA22FF; font-weight: bold\">=</span><span style=\"color: #008000\">48</span>, kernel_size<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1</span>]) <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">48</span>, <span style=\"color: #008000\">12</span>, <span style=\"color: #008000\">20</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=/</span>conv22<span style=\"color: #AA22FF; font-weight: bold\">/</span>Conv:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>;\n",
       "  <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">9</span> <span style=\"color: #AA22FF; font-weight: bold\">=</span> llvm_special<span style=\"color: #AA22FF; font-weight: bold\">.</span>det_class_predict(<span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">3</span>, <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">4</span>, <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">5</span>, __dict__<span style=\"color: #AA22FF; font-weight: bold\">=</span>{<span style=\"color: #BA2121\">&quot;class_num&quot;</span><span style=\"color: #AA22FF; font-weight: bold\">=</span><span style=\"color: #008000\">3</span>});\n",
       "  <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">10</span> <span style=\"color: #AA22FF; font-weight: bold\">=</span> llvm_special<span style=\"color: #AA22FF; font-weight: bold\">.</span>det_bbox_predict(<span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">6</span>, <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">7</span>, <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">8</span>);\n",
       "  (<span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">9</span>, <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">10</span>) <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>(Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">61440</span>, <span style=\"color: #008000\">3</span>), float32], Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">46080</span>, <span style=\"color: #008000\">4</span>), float32]) <span style=\"color: #AA22FF; font-weight: bold\">*/</span>\n",
       "}\n",
       "</pre></div>\n"
      ],
      "text/plain": [
       "<IPython.core.display.HTML object>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "mod.show()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 12,
   "metadata": {},
   "outputs": [
    {
     "name": "stderr",
     "output_type": "stream",
     "text": [
      "One or more operators have not been tuned. Please tune your model for better performance. Use DEBUG logging level to see more details.\n"
     ]
    }
   ],
   "source": [
    "target = \"llvm\"\n",
    "dev = tvm.device(target, 0)\n",
    "with tvm.transform.PassContext(opt_level=3):\n",
    "    lib = relay.build(mod, target)\n",
    "func = lib[lib.libmod_name]\n",
    "module = tvm.contrib.graph_executor.GraphModule(func(dev))\n",
    "module.run(**{config.name: data_np})\n",
    "outputs = [module.get_output(k).numpy() for k in range(2)]"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 13,
   "metadata": {},
   "outputs": [],
   "source": [
    "with tvm.transform.PassContext(opt_level=3):\n",
    "    lib = relay.build(origin_mod, target)\n",
    "func = lib[lib.libmod_name]\n",
    "module = tvm.contrib.graph_executor.GraphModule(func(dev))\n",
    "module.run(**{config.name: data_np})\n",
    "origon_outputs = [module.get_output(k).numpy() for k in range(2)]"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 14,
   "metadata": {},
   "outputs": [],
   "source": [
    "with torch.no_grad():\n",
    "    torch_outputs = [o.numpy() for o in model(torch.from_numpy(data_np))]"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 15,
   "metadata": {},
   "outputs": [],
   "source": [
    "[\n",
    "    np.testing.assert_allclose(torch_output, origon_output, rtol=1e-07, atol=1e-05)\n",
    "    for torch_output, origon_output in zip(torch_outputs, origon_outputs)\n",
    "]\n",
    "[\n",
    "    np.testing.assert_allclose(torch_output, output, rtol=1e-07, atol=1e-05)\n",
    "    for torch_output, output in zip(torch_outputs, outputs)\n",
    "];"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 16,
   "metadata": {},
   "outputs": [],
   "source": [
    "from tvm.relay.op import op as _op\n",
    "from special_op.utils import schedule_special_op"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 18,
   "metadata": {},
   "outputs": [],
   "source": [
    "# from tvm import te, topi\n",
    "# from tvm.relay.op import strategy as _strategy\n",
    "# from tvm.relay.op.op import OpPattern, OpStrategy\n",
    "\n",
    "# def det_bbox_predict_compute(attrs, inputs, out_type):\n",
    "#     \"\"\"det_bbox_predict Relay 计算\"\"\"\n",
    "#     assert len(inputs) == 3, \"输入参数数量不为 3\"\n",
    "#     x0, x1, x2 = inputs\n",
    "#     bbox_size = 4\n",
    "#     x0 = topi.transpose(x0, [0, 2, 3, 1])\n",
    "#     x0 = topi.reshape(x0, [1, x0.shape[1]*x0.shape[2]*x0.shape[3]//bbox_size, bbox_size])\n",
    "#     x1 = topi.transpose(x1, [0, 2, 3, 1])\n",
    "#     x1 = topi.reshape(x1, [1, x1.shape[1]*x1.shape[2]*x1.shape[3]//bbox_size, bbox_size])\n",
    "#     x2 = topi.transpose(x2, [0, 2, 3, 1])\n",
    "#     x2 = topi.reshape(x2, [1, x2.shape[1]*x2.shape[2]*x2.shape[3]//bbox_size, bbox_size])\n",
    "#     x = topi.concatenate([x0, x1, x2], axis=1)\n",
    "#     print(x)\n",
    "#     return [x]\n",
    "\n",
    "\n",
    "# op_name = \"vta_special.det_bbox_predict\"\n",
    "# def det_bbox_predict_strategy_vta(attrs, inputs, out_type, target):\n",
    "#     strategy = OpStrategy()\n",
    "#     strategy.add_implementation(\n",
    "#         det_bbox_predict_compute,\n",
    "#         schedule_special_op,\n",
    "#         name=f\"{op_name}.llvm\",\n",
    "#     )\n",
    "#     return strategy\n",
    "\n",
    "\n",
    "# _op.get(op_name).get_attr(\"FTVMStrategy\").register(det_bbox_predict_strategy_vta, \"llvm\", allow_override=True)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 19,
   "metadata": {},
   "outputs": [],
   "source": [
    "with tvm.transform.PassContext(opt_level=3):\n",
    "    lib = relay.build(mod, \"llvm\")"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "\n",
    "# func = lib[lib.libmod_name]\n",
    "# module = tvm.contrib.graph_executor.GraphModule(func(tvm.ext_dev(0)))\n",
    "# module.run(**{config.name: data_np})\n",
    "# outputs = [module.get_output(k).numpy() for k in range(2)]"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 24,
   "metadata": {},
   "outputs": [],
   "source": [
    "from tvm.relay.dataflow_pattern import wildcard, is_constant, is_op\n",
    "# import logging\n",
    "\n",
    "def is_QPartitionExpr(op):\n",
    "    r = is_op(\"annotation.cast_hint\")(op)\n",
    "    r = is_op(\"annotation.stop_fusion\")(r)\n",
    "    return r\n",
    "\n",
    "def debug_partition(func):\n",
    "    def is_QPartitionExpr():\n",
    "        # logging.debug(f\"enter {func.__name__}()\")\n",
    "        r = func()\n",
    "        r = is_op(\"annotation.cast_hint\")(r) | r\n",
    "        # r = is_op(\"annotation.stop_fusion\")(r) | r\n",
    "        return r\n",
    "    return is_QPartitionExpr\n",
    "\n",
    "@debug_partition\n",
    "def make_conv_add_squeeze_pattern():\n",
    "    x = wildcard()\n",
    "    w = wildcard()\n",
    "    bias = wildcard()\n",
    "    x_ = is_op(\"relay.op.annotation.simulated_quantize\")(x, is_constant(), is_constant(), is_constant())\n",
    "    x_ = x_ | x\n",
    "    w_ = is_op(\"relay.op.annotation.simulated_quantize\")(w, is_constant(), is_constant(), is_constant())\n",
    "    w_ = w_ | w\n",
    "    bias_ = is_op(\"relay.op.annotation.simulated_quantize\")(bias, is_constant(), is_constant(), is_constant()) | bias\n",
    "    conv_node = is_op(\"nn.conv2d\")(x_, w_)\n",
    "    r = is_op(\"add\")(conv_node, bias_)\n",
    "    r = is_op(\"relay.op.annotation.simulated_quantize\")(r, is_constant(), is_constant(), is_constant()) | r\n",
    "    r = is_op(\"squeeze\")(r)\n",
    "    r = is_op(\"relay.op.annotation.simulated_quantize\")(r, is_constant(), is_constant(), is_constant()) | r\n",
    "    return r\n",
    "\n",
    "@debug_partition\n",
    "def make_conv_add_relu_max_pool2d_pattern():\n",
    "    x = wildcard()\n",
    "    w = wildcard()\n",
    "    bias = wildcard()\n",
    "    # x_ = is_op(\"relay.op.annotation.simulated_quantize\")(x, is_constant(), is_constant(), is_constant()) | x\n",
    "    # w_ = is_op(\"relay.op.annotation.simulated_quantize\")(w, is_constant(), is_constant(), is_constant()) | w\n",
    "    bias_ = is_op(\"relay.op.annotation.simulated_quantize\")(bias, is_constant(), is_constant(), is_constant()) | bias\n",
    "    conv_node = is_op(\"nn.conv2d\")(x, w)\n",
    "    r = is_op(\"add\")(conv_node, bias_) | conv_node\n",
    "    r = is_op(\"nn.relu\")(r) | is_op(\"nn.prelu\")(r, wildcard())\n",
    "    r = is_op(\"relay.op.annotation.simulated_quantize\")(r, is_constant(), is_constant(), is_constant()) | r\n",
    "    xx = is_op(\"annotation.cast_hint\")(r)\n",
    "    xx = is_op(\"annotation.stop_fusion\")(xx)\n",
    "    # xx = is_op(\"nn.max_pool2d\")(xx)\n",
    "    r = is_op(\"nn.max_pool2d\")(xx|r).has_attr({\n",
    "        \"padding\": [0, 0, 0, 0],\n",
    "        # \"ceil_mode\": False,\n",
    "    })\n",
    "    r = is_op(\"relay.op.annotation.simulated_quantize\")(r, is_constant(), is_constant(), is_constant()) | r\n",
    "    return r\n",
    "\n",
    "@debug_partition\n",
    "def make_conv_add_activate_pattern():\n",
    "    r\"\"\"Create a pattern to match the following graph.\n",
    "\n",
    "    conv2d\n",
    "        |\n",
    "    add\n",
    "        |\n",
    "        (relu|relu6|prelu|sigmoid|relux)\n",
    "    \"\"\"\n",
    "    x = wildcard()\n",
    "    w = wildcard()\n",
    "    bias = wildcard()\n",
    "    alpha = wildcard()\n",
    "    # x_ = is_op(\"relay.op.annotation.simulated_quantize\")(x, is_constant(), is_constant(), is_constant()) | x\n",
    "    # w_ = is_op(\"relay.op.annotation.simulated_quantize\")(w, is_constant(), is_constant(), is_constant()) | w\n",
    "    bias_ = is_op(\"relay.op.annotation.simulated_quantize\")(bias, is_constant(), is_constant(), is_constant()) | bias\n",
    "    # alpha_ = is_op(\"relay.op.annotation.simulated_quantize\")(alpha, is_constant(), is_constant(), is_constant()) | alpha\n",
    "    conv_node = is_op(\"nn.conv2d\")(x, w)\n",
    "    r = is_op(\"add\")(conv_node, bias_) | is_op(\"nn.bias_add\")(conv_node, bias_) | conv_node\n",
    "    \n",
    "    # 激活函数\n",
    "    r1 = r.optional(lambda x: is_op(\"nn.relu\")(x))\n",
    "    r2 = r.optional(lambda x: is_op(\"clip\")(x)) # relu6\n",
    "    r3 = r.optional(lambda x: is_op(\"nn.prelu\")(x, alpha)) # prelu\n",
    "    r4 = r.optional(lambda x: is_op(\"sigmoid\")(x)) # sigmoid\n",
    "    # r5 = r.optional(lambda x: is_op(\"nn.relux\")(x, alpha)) # relux\n",
    "    # r6 = r.optional(lambda x: is_op(\"silu\")(x)) # silu\n",
    "    # r7 = r.optional(lambda x: is_op(\"hard_sigmoid\")(x)) # hard_sigmoid\n",
    "    # r8 = r.optional(lambda x: is_op(\"hard_swish\")(x)) # hard_swish\n",
    "    r = r1 | r2 | r3 | r4\n",
    "\n",
    "    r = is_op(\"relay.op.annotation.simulated_quantize\")(r, is_constant(), is_constant(), is_constant()) | r\n",
    "    return r\n",
    "\n",
    "@debug_partition\n",
    "def make_conv2d_transpose_add_activate_pattern():\n",
    "    r\"\"\"Create a pattern to match the following graph.\n",
    "\n",
    "    conv2d_transpose\n",
    "        |\n",
    "    add\n",
    "    \"\"\"\n",
    "    x = wildcard()\n",
    "    w = wildcard()\n",
    "    bias = wildcard()\n",
    "    alpha = wildcard()\n",
    "    alpha_ = is_op(\"relay.op.annotation.simulated_quantize\")(alpha, is_constant(), is_constant(), is_constant()) | alpha\n",
    "    # x_ = is_op(\"relay.op.annotation.simulated_quantize\")(x, is_constant(), is_constant(), is_constant()) | x\n",
    "    # w_ = is_op(\"relay.op.annotation.simulated_quantize\")(w, is_constant(), is_constant(), is_constant()) | w\n",
    "    bias_ = is_op(\"relay.op.annotation.simulated_quantize\")(bias, is_constant(), is_constant(), is_constant()) | bias\n",
    "    r = is_op(\"nn.conv2d_transpose\")(x, w)\n",
    "    r = is_op(\"add\")(r, bias_) | r\n",
    "    \n",
    "    # 激活函数\n",
    "    r1 = r.optional(lambda x: is_op(\"nn.relu\")(x))\n",
    "    r2 = r.optional(lambda x: is_op(\"clip\")(x)) # relu6\n",
    "    r3 = r.optional(lambda x: is_op(\"nn.prelu\")(x, alpha)) # prelu\n",
    "    r4 = r.optional(lambda x: is_op(\"sigmoid\")(x)) # sigmoid\n",
    "    r = r1 | r2 | r3 | r4\n",
    "\n",
    "    r = is_op(\"relay.op.annotation.simulated_quantize\")(r, is_constant(), is_constant(), is_constant()) | r\n",
    "    # r = is_QPartitionExpr(r) | r\n",
    "    # r = is_op(\"strided_slice\")(r)\n",
    "    return r\n",
    "\n",
    "@debug_partition\n",
    "def make_max_pool2d_pattern():\n",
    "    x = wildcard()\n",
    "    r = is_op(\"nn.max_pool2d\")(x)\n",
    "    r = is_op(\"relay.op.annotation.simulated_quantize\")(r, is_constant(), is_constant(), is_constant()) | r\n",
    "    return r\n",
    "\n",
    "@debug_partition\n",
    "def make_strided_slice_pattern():\n",
    "    x = wildcard()\n",
    "    r = is_op(\"strided_slice\")(x)\n",
    "    return r\n",
    "\n",
    "@debug_partition\n",
    "def make_concat_pattern():\n",
    "    x = wildcard()\n",
    "    r = is_op(\"concatenate\")(x)\n",
    "    r = is_op(\"relay.op.annotation.simulated_quantize\")(r, is_constant(), is_constant(), is_constant()) | r\n",
    "    return r\n",
    "\n",
    "@debug_partition\n",
    "def make_avg_pool2d_pattern():\n",
    "    x = wildcard()\n",
    "    r = is_op(\"nn.avg_pool2d\")(x) \n",
    "    r = is_op(\"relay.op.annotation.simulated_quantize\")(r, is_constant(), is_constant(), is_constant()) | r\n",
    "    return r\n",
    "\n",
    "@debug_partition\n",
    "def make_adaptive_avg_pool2d_pattern():\n",
    "    x = wildcard()\n",
    "    r = is_op(\"nn.adaptive_avg_pool2d\")(x) \n",
    "    r = is_op(\"relay.op.annotation.simulated_quantize\")(r, is_constant(), is_constant(), is_constant()) | r\n",
    "    return r\n",
    "\n",
    "@debug_partition\n",
    "def make_dense_add_pattern():\n",
    "    r\"\"\"Create a pattern to match the following graph.\n",
    "\n",
    "      nn.dense\n",
    "        |\n",
    "       add\n",
    "    \"\"\"\n",
    "    x = wildcard()\n",
    "    y = wildcard()\n",
    "    w = wildcard()\n",
    "    # x_ = is_op(\"relay.op.annotation.simulated_quantize\")(x, is_constant(), is_constant(), is_constant()) | x\n",
    "    # y_ = is_op(\"relay.op.annotation.simulated_quantize\")(y, is_constant(), is_constant(), is_constant()) | y\n",
    "    w_ = is_op(\"relay.op.annotation.simulated_quantize\")(w, is_constant(), is_constant(), is_constant()) | w\n",
    "    node = is_op(\"nn.dense\")(x, y)\n",
    "    r = is_op(\"add\")(node, w_) | node\n",
    "    r = is_op(\"relay.op.annotation.simulated_quantize\")(r, is_constant(), is_constant(), is_constant()) | r\n",
    "    return r\n",
    "\n",
    "@debug_partition\n",
    "def make_add_add_pattern():\n",
    "    \"\"\"\n",
    "    z = x + y\n",
    "    z2 = z + c1\n",
    "    其中 c1 为常量\n",
    "    \"\"\"\n",
    "    c1 = is_constant()\n",
    "    c1_ = is_op(\"relay.op.annotation.simulated_quantize\")(c1, is_constant(), is_constant(), is_constant()) | c1\n",
    "\n",
    "    r = wildcard() + wildcard()\n",
    "    r = is_op(\"relay.op.annotation.simulated_quantize\")(r, is_constant(), is_constant(), is_constant()) | r\n",
    "    r = r + c1_\n",
    "    r = is_op(\"relay.op.annotation.simulated_quantize\")(r, is_constant(), is_constant(), is_constant()) | r\n",
    "    return r\n",
    "\n",
    "@debug_partition\n",
    "def make_multiply_add_pattern():\n",
    "    \"\"\"\n",
    "    z = x * y\n",
    "    z2 = z + c1\n",
    "    \"\"\"\n",
    "    c1 = is_constant()\n",
    "    c1_ = is_op(\"relay.op.annotation.simulated_quantize\")(c1, is_constant(), is_constant(), is_constant()) | c1\n",
    "\n",
    "    r = wildcard() * wildcard()\n",
    "    r = is_op(\"relay.op.annotation.simulated_quantize\")(r, is_constant(), is_constant(), is_constant()) | r\n",
    "    r = r + c1_\n",
    "    r = is_op(\"relay.op.annotation.simulated_quantize\")(r, is_constant(), is_constant(), is_constant()) | r\n",
    "    return r\n",
    "\n",
    "@debug_partition\n",
    "def make_add_multiply_add_pattern():\n",
    "    \"\"\"\n",
    "    z = x + y\n",
    "    z2 = z * c1 + c2\n",
    "    其中 c1，c2 为常量\n",
    "    \"\"\"\n",
    "    c1 = is_constant()\n",
    "    c2 = is_constant()\n",
    "    c1_ = is_op(\"relay.op.annotation.simulated_quantize\")(c1, is_constant(), is_constant(), is_constant()) | c1\n",
    "    c2_ = is_op(\"relay.op.annotation.simulated_quantize\")(c2, is_constant(), is_constant(), is_constant()) | c2\n",
    "\n",
    "    r = wildcard() + wildcard()\n",
    "    r = is_op(\"relay.op.annotation.simulated_quantize\")(r, is_constant(), is_constant(), is_constant()) | r\n",
    "    r = r * c1_\n",
    "    r = is_op(\"relay.op.annotation.simulated_quantize\")(r, is_constant(), is_constant(), is_constant()) | r\n",
    "    r = r + c2_\n",
    "    r = is_op(\"relay.op.annotation.simulated_quantize\")(r, is_constant(), is_constant(), is_constant()) | r\n",
    "    return r\n",
    "\n",
    "@debug_partition\n",
    "def make_add_pattern():\n",
    "    r\"\"\"Create a pattern to match the following graph.\n",
    "\n",
    "      add\n",
    "        |\n",
    "      relu|relu6\n",
    "    \"\"\"\n",
    "    r = wildcard() + wildcard()\n",
    "    r1 = r.optional(lambda x: is_op(\"nn.relu\")(x))\n",
    "    r2 = r.optional(lambda x: is_op(\"clip\")(x)) # relu6\n",
    "    r = r1 | r2\n",
    "    r = is_op(\"relay.op.annotation.simulated_quantize\")(r, is_constant(), is_constant(), is_constant()) | r\n",
    "    return r\n",
    "\n",
    "@debug_partition\n",
    "def make_multiply_pattern():\n",
    "    r\"\"\"Create a pattern to match the following graph.\n",
    "\n",
    "      multiply\n",
    "        |\n",
    "      relu|relu6\n",
    "    \"\"\"\n",
    "    r = wildcard() * wildcard()\n",
    "    r1 = r.optional(lambda x: is_op(\"nn.relu\")(x))\n",
    "    r2 = r.optional(lambda x: is_op(\"clip\")(x)) # relu6\n",
    "    r = r1 | r2\n",
    "    r = is_op(\"relay.op.annotation.simulated_quantize\")(r, is_constant(), is_constant(), is_constant()) | r\n",
    "    return r\n",
    "\n",
    "def make_reshape_squeeze_pattern():\n",
    "    x = wildcard()\n",
    "    r = is_op(\"reshape\")(x)\n",
    "    r = is_op(\"squeeze\")(r)\n",
    "    r = is_op(\"relay.op.annotation.simulated_quantize\")(r, is_constant(), is_constant(), is_constant()) | r\n",
    "    return r\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 25,
   "metadata": {},
   "outputs": [],
   "source": [
    "# 配置融合规则\n",
    "compiler_name = \"vta_special\"\n",
    "pattern_table = [\n",
    "    # (f\"{compiler_name}.concat\", make_concat_pattern()),\n",
    "    (f\"{compiler_name}.conv_add_squeeze\", make_conv_add_squeeze_pattern()), # mobilenet_v2_tf 最后一层\n",
    "    (f\"{compiler_name}.conv_add_relu_max_pool2d\", make_conv_add_relu_max_pool2d_pattern()),\n",
    "    (f\"{compiler_name}.conv2d_transpose_add_activate\", make_conv2d_transpose_add_activate_pattern()),\n",
    "    (f\"{compiler_name}.conv_add_activate\", make_conv_add_activate_pattern()),\n",
    "    (f\"{compiler_name}.max_pool2d\", make_max_pool2d_pattern()),\n",
    "    (f\"{compiler_name}.dense_add\", make_dense_add_pattern()),\n",
    "    (f\"{compiler_name}.adaptive_avg_pool2d\", make_adaptive_avg_pool2d_pattern()),\n",
    "    (f\"{compiler_name}.avg_pool2dd\", make_avg_pool2d_pattern()),\n",
    "    (f\"{compiler_name}.add_multiply_add\", make_add_multiply_add_pattern()), # kr_karen\n",
    "    (f\"{compiler_name}.add_add\", make_add_add_pattern()),\n",
    "    (f\"{compiler_name}.multiply_add\", make_multiply_add_pattern()),\n",
    "    (f\"{compiler_name}.add\", make_add_pattern()),\n",
    "    (f\"{compiler_name}.multiply\", make_multiply_pattern()),\n",
    "    # (\"det_class_predict_temp\", make_det_class_predict_temp_pattern()),\n",
    "    # (\"det_bbox_predict_temp\",make_det_bbox_predict_temp_pattern())\n",
    "    # (f\"{compiler_name}.strided_slice\", make_strided_slice_pattern()),\n",
    "]\n",
    "merge_passes = tvm.transform.Sequential([\n",
    "    relay.transform.InferType(),\n",
    "    relay.transform.MergeComposite(pattern_table),\n",
    "])"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 26,
   "metadata": {},
   "outputs": [],
   "source": [
    "run_mod = deepcopy(mod)\n",
    "with tvm.transform.PassContext(opt_level=3):\n",
    "    partition_mod_t = merge_passes(run_mod)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 27,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "def @main(%data: Tensor[(1, 3, 48, 80), float32] /* ty=Tensor[(1, 3, 48, 80), float32] span=/conv/Conv.data:0:0 */) -> (Tensor[(1, 61440, 3), float32], Tensor[(1, 46080, 4), float32]) {\n",
      "  %0 = fn (%FunctionVar_8_0: Tensor[(1, 3, 48, 80), float32] /* ty=Tensor[(1, 3, 48, 80), float32] */, %FunctionVar_8_1: Tensor[(64, 3, 1, 1), float32] /* ty=Tensor[(64, 3, 1, 1), float32] */, PartitionedFromPattern=\"nn.conv2d_\", Composite=\"vta_special.conv_add_activate\") -> Tensor[(1, 64, 48, 80), float32] {\n",
      "    nn.conv2d(%FunctionVar_8_0, %FunctionVar_8_1, padding=[0, 0, 0, 0], channels=64, kernel_size=[1, 1]) /* ty=Tensor[(1, 64, 48, 80), float32] span=/conv/Conv:0:0 */\n",
      "  } /* ty=fn (Tensor[(1, 3, 48, 80), float32], Tensor[(64, 3, 1, 1), float32]) -> Tensor[(1, 64, 48, 80), float32] */;\n",
      "  %1 = %0(%data, meta[relay.Constant][0] /* ty=Tensor[(64, 3, 1, 1), float32] span=/conv/Conv.conv.weight:0:0 */) /* ty=Tensor[(1, 64, 48, 80), float32] */;\n",
      "  %2 = fn (%FunctionVar_7_0: Tensor[(1, 64, 48, 80), float32] /* ty=Tensor[(1, 64, 48, 80), float32] */, %FunctionVar_7_1: Tensor[(36, 64, 1, 1), float32] /* ty=Tensor[(36, 64, 1, 1), float32] */, PartitionedFromPattern=\"nn.conv2d_\", Composite=\"vta_special.conv_add_activate\") -> Tensor[(1, 36, 48, 80), float32] {\n",
      "    nn.conv2d(%FunctionVar_7_0, %FunctionVar_7_1, padding=[0, 0, 0, 0], channels=36, kernel_size=[1, 1]) /* ty=Tensor[(1, 36, 48, 80), float32] span=/conv0/Conv:0:0 */\n",
      "  } /* ty=fn (Tensor[(1, 64, 48, 80), float32], Tensor[(36, 64, 1, 1), float32]) -> Tensor[(1, 36, 48, 80), float32] */;\n",
      "  %3 = fn (%FunctionVar_6_0: Tensor[(1, 64, 48, 80), float32] /* ty=Tensor[(1, 64, 48, 80), float32] */, %FunctionVar_6_1: Tensor[(64, 64, 1, 1), float32] /* ty=Tensor[(64, 64, 1, 1), float32] */, PartitionedFromPattern=\"nn.conv2d_\", Composite=\"vta_special.conv_add_activate\") -> Tensor[(1, 64, 24, 40), float32] {\n",
      "    nn.conv2d(%FunctionVar_6_0, %FunctionVar_6_1, strides=[2, 2], padding=[0, 0, 0, 0], channels=64, kernel_size=[1, 1]) /* ty=Tensor[(1, 64, 24, 40), float32] span=/resize_2/Conv:0:0 */\n",
      "  } /* ty=fn (Tensor[(1, 64, 48, 80), float32], Tensor[(64, 64, 1, 1), float32]) -> Tensor[(1, 64, 24, 40), float32] */;\n",
      "  %4 = %3(%1, meta[relay.Constant][2] /* ty=Tensor[(64, 64, 1, 1), float32] span=/resize_2/Conv.resize_2.weight:0:0 */) /* ty=Tensor[(1, 64, 24, 40), float32] */;\n",
      "  %5 = fn (%FunctionVar_5_0: Tensor[(1, 64, 24, 40), float32] /* ty=Tensor[(1, 64, 24, 40), float32] */, %FunctionVar_5_1: Tensor[(36, 64, 1, 1), float32] /* ty=Tensor[(36, 64, 1, 1), float32] */, PartitionedFromPattern=\"nn.conv2d_\", Composite=\"vta_special.conv_add_activate\") -> Tensor[(1, 36, 24, 40), float32] {\n",
      "    nn.conv2d(%FunctionVar_5_0, %FunctionVar_5_1, padding=[0, 0, 0, 0], channels=36, kernel_size=[1, 1]) /* ty=Tensor[(1, 36, 24, 40), float32] span=/conv1/Conv:0:0 */\n",
      "  } /* ty=fn (Tensor[(1, 64, 24, 40), float32], Tensor[(36, 64, 1, 1), float32]) -> Tensor[(1, 36, 24, 40), float32] */;\n",
      "  %6 = fn (%FunctionVar_4_0: Tensor[(1, 64, 48, 80), float32] /* ty=Tensor[(1, 64, 48, 80), float32] */, %FunctionVar_4_1: Tensor[(64, 64, 1, 1), float32] /* ty=Tensor[(64, 64, 1, 1), float32] */, PartitionedFromPattern=\"nn.conv2d_\", Composite=\"vta_special.conv_add_activate\") -> Tensor[(1, 64, 12, 20), float32] {\n",
      "    nn.conv2d(%FunctionVar_4_0, %FunctionVar_4_1, strides=[4, 4], padding=[0, 0, 0, 0], channels=64, kernel_size=[1, 1]) /* ty=Tensor[(1, 64, 12, 20), float32] span=/resize_4/Conv:0:0 */\n",
      "  } /* ty=fn (Tensor[(1, 64, 48, 80), float32], Tensor[(64, 64, 1, 1), float32]) -> Tensor[(1, 64, 12, 20), float32] */;\n",
      "  %7 = %6(%1, meta[relay.Constant][4] /* ty=Tensor[(64, 64, 1, 1), float32] span=/resize_4/Conv.resize_4.weight:0:0 */) /* ty=Tensor[(1, 64, 12, 20), float32] */;\n",
      "  %8 = fn (%FunctionVar_3_0: Tensor[(1, 64, 12, 20), float32] /* ty=Tensor[(1, 64, 12, 20), float32] */, %FunctionVar_3_1: Tensor[(48, 64, 1, 1), float32] /* ty=Tensor[(48, 64, 1, 1), float32] */, PartitionedFromPattern=\"nn.conv2d_\", Composite=\"vta_special.conv_add_activate\") -> Tensor[(1, 48, 12, 20), float32] {\n",
      "    nn.conv2d(%FunctionVar_3_0, %FunctionVar_3_1, padding=[0, 0, 0, 0], channels=48, kernel_size=[1, 1]) /* ty=Tensor[(1, 48, 12, 20), float32] span=/conv2/Conv:0:0 */\n",
      "  } /* ty=fn (Tensor[(1, 64, 12, 20), float32], Tensor[(48, 64, 1, 1), float32]) -> Tensor[(1, 48, 12, 20), float32] */;\n",
      "  %9 = %2(%1, meta[relay.Constant][1] /* ty=Tensor[(36, 64, 1, 1), float32] span=/conv0/Conv.conv0.weight:0:0 */) /* ty=Tensor[(1, 36, 48, 80), float32] */;\n",
      "  %10 = %5(%4, meta[relay.Constant][3] /* ty=Tensor[(36, 64, 1, 1), float32] span=/conv1/Conv.conv1.weight:0:0 */) /* ty=Tensor[(1, 36, 24, 40), float32] */;\n",
      "  %11 = %8(%7, meta[relay.Constant][5] /* ty=Tensor[(48, 64, 1, 1), float32] span=/conv2/Conv.conv2.weight:0:0 */) /* ty=Tensor[(1, 48, 12, 20), float32] */;\n",
      "  %12 = fn (%FunctionVar_2_0: Tensor[(1, 64, 48, 80), float32] /* ty=Tensor[(1, 64, 48, 80), float32] */, %FunctionVar_2_1: Tensor[(36, 64, 1, 1), float32] /* ty=Tensor[(36, 64, 1, 1), float32] */, PartitionedFromPattern=\"nn.conv2d_\", Composite=\"vta_special.conv_add_activate\") -> Tensor[(1, 36, 48, 80), float32] {\n",
      "    nn.conv2d(%FunctionVar_2_0, %FunctionVar_2_1, padding=[0, 0, 0, 0], channels=36, kernel_size=[1, 1]) /* ty=Tensor[(1, 36, 48, 80), float32] span=/conv00/Conv:0:0 */\n",
      "  } /* ty=fn (Tensor[(1, 64, 48, 80), float32], Tensor[(36, 64, 1, 1), float32]) -> Tensor[(1, 36, 48, 80), float32] */;\n",
      "  %13 = fn (%FunctionVar_1_0: Tensor[(1, 64, 24, 40), float32] /* ty=Tensor[(1, 64, 24, 40), float32] */, %FunctionVar_1_1: Tensor[(36, 64, 1, 1), float32] /* ty=Tensor[(36, 64, 1, 1), float32] */, PartitionedFromPattern=\"nn.conv2d_\", Composite=\"vta_special.conv_add_activate\") -> Tensor[(1, 36, 24, 40), float32] {\n",
      "    nn.conv2d(%FunctionVar_1_0, %FunctionVar_1_1, padding=[0, 0, 0, 0], channels=36, kernel_size=[1, 1]) /* ty=Tensor[(1, 36, 24, 40), float32] span=/conv11/Conv:0:0 */\n",
      "  } /* ty=fn (Tensor[(1, 64, 24, 40), float32], Tensor[(36, 64, 1, 1), float32]) -> Tensor[(1, 36, 24, 40), float32] */;\n",
      "  %14 = fn (%FunctionVar_0_0: Tensor[(1, 64, 12, 20), float32] /* ty=Tensor[(1, 64, 12, 20), float32] */, %FunctionVar_0_1: Tensor[(48, 64, 1, 1), float32] /* ty=Tensor[(48, 64, 1, 1), float32] */, PartitionedFromPattern=\"nn.conv2d_\", Composite=\"vta_special.conv_add_activate\") -> Tensor[(1, 48, 12, 20), float32] {\n",
      "    nn.conv2d(%FunctionVar_0_0, %FunctionVar_0_1, padding=[0, 0, 0, 0], channels=48, kernel_size=[1, 1]) /* ty=Tensor[(1, 48, 12, 20), float32] span=/conv22/Conv:0:0 */\n",
      "  } /* ty=fn (Tensor[(1, 64, 12, 20), float32], Tensor[(48, 64, 1, 1), float32]) -> Tensor[(1, 48, 12, 20), float32] */;\n",
      "  %15 = %12(%1, meta[relay.Constant][6] /* ty=Tensor[(36, 64, 1, 1), float32] span=/conv00/Conv.conv00.weight:0:0 */) /* ty=Tensor[(1, 36, 48, 80), float32] */;\n",
      "  %16 = %13(%4, meta[relay.Constant][7] /* ty=Tensor[(36, 64, 1, 1), float32] span=/conv11/Conv.conv11.weight:0:0 */) /* ty=Tensor[(1, 36, 24, 40), float32] */;\n",
      "  %17 = %14(%7, meta[relay.Constant][8] /* ty=Tensor[(48, 64, 1, 1), float32] span=/conv22/Conv.conv22.weight:0:0 */) /* ty=Tensor[(1, 48, 12, 20), float32] */;\n",
      "  %18 = llvm_special.det_class_predict(%9, %10, %11, __dict__={\"class_num\"=3}) /* ty=Tensor[(1, 61440, 3), float32] */;\n",
      "  %19 = llvm_special.det_bbox_predict(%15, %16, %17) /* ty=Tensor[(1, 46080, 4), float32] */;\n",
      "  (%18, %19) /* ty=(Tensor[(1, 61440, 3), float32], Tensor[(1, 46080, 4), float32]) */\n",
      "}\n",
      "\n",
      "\n"
     ]
    }
   ],
   "source": [
    "print(partition_mod_t)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": []
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "py312x",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.12.3"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
